// Copyright 2006-2007 all rights reserved; see LICENSE file for BSD-style license

package edu.berkeley.sbp;
import edu.berkeley.sbp.*;
import edu.berkeley.sbp.util.*;
import edu.berkeley.sbp.Parser.Table.*;
import edu.berkeley.sbp.Sequence.Position;
import edu.berkeley.sbp.Sequence.Pos;
import java.io.*;
import java.util.*;
import java.lang.reflect.*;

/** a node in the GSS */
final class Node
    implements Invokable<Pos, Result>,
               IntegerMappable,
               GraphViz.ToGraphViz,
               Iterable<Result> {

    /** which GSS.Phase this Node belongs to */
    public GSS.Phase phase() { return phase; }
    public Iterator<Result> iterator() { return results.iterator(); }
    public Parser.Table.State state() { return state; }

    public int toInt() { return idx; }

    boolean destroyed = false;

    public void check() {
        if (destroyed) return;
        boolean dead = true;
        // - all nodes must have a parent
        // - non-doomed nodes must either:
        //      - be on the frontier or
        //      - have a non-doomed node closer to the frontier than themself
        if (phase.isFrontier()) dead = false;
        else for(Result r : children)
                 if (state.doomed) { if (r.usedByAnyNode()) { dead = false; break; } }
                 else              { if (r.usedByNonDoomedNode()) { dead = false; break; } }
        dead |= results.size()==0;
        if (!dead) return;
        destroyed = true;
        while(children.size()>0)
            for(Result r : children) {
                children.remove(r);
                r.destroy();
                break;
            }
        while(results.size()>0)
            for(Result r : results) {
                results.remove(r);
                r.removeChild(this);
                break;
            }
        results = null;
        children = null;
    }

    //////////////////////////////////////////////////////////////////////

    private static int node_idx = 0;
    private final int idx = node_idx++;

    private final GSS.Phase phase;
    private final Parser.Table.State state;
    private final boolean fromEmptyReduction;
    //private       FastSet<Result> results = new FastSet<Result>();
    private       HashSet<Result> results = new HashSet<Result>();
    private       HashSet<Result> children = new HashSet<Result>();

    public final void invoke(Pos r, Result only) {
        boolean emptyProductions = only==null;
        if (emptyProductions != (r.numPops()==0)) return;
        if (r.numPops()!=0)  reduce(r, r.numPops()-1, phase(), only);
        else {
            Input.Region region = phase().getLocation().createRegion(phase().getLocation());
            new Result(r.rewrite(region, phase().parser().cache()), this, r, phase());
        }
    }

    private void reduce(Pos r, int pos, GSS.Phase target, Result only) {
        Forest[] holder = r.holder;
        Forest old = holder[pos];
        if (results==null) return;   // FIXME: this should not happen
        for(Result res : results)
            if (only == null || res == only) {
                Node child = res.parent();
                holder[pos] = res.getForest();
                if (pos>0)  child.reduce(r, pos-1, target, null);
                else {
                    Input.Region region = child.phase().getLocation().createRegion(target.getLocation());
                    new Reduction(child, r, r.rewrite(region, phase().parser().cache()), target);
                }
            }
        holder[pos] = old;
    }

    Node(GSS.Phase phase, Result result, State state, boolean fromEmptyReduction) {
        this.phase = phase;
        this.state = state;
        this.fromEmptyReduction = fromEmptyReduction;
        if (phase.hash.get(state, result.phase()) != null) throw new Error("severe problem!");
        phase.hash.put(state, result.phase(), this);
        addResult(result);
        state.invokeEpsilonReductions(phase().token, this);
    }

    // Add/Remove Children/Results //////////////////////////////////////////////////////////////////////////////

    public void removeChild(Result child)   {
        if (children==null) return;
        children.remove(child);
        check();
    }
    public void removeResult(Result result) {
        if (results==null) return;
        results.remove(result);
        check();
    }
    public void addChild(Result child)      {
        if (children==null) return;   // FIXME: this should not happen
        children.add(child);
    }
    public void addResult(Result r) {
        if (results.contains(r)) return;
        results.add(r);
        r.addChild(this);
        if (!fromEmptyReduction) state.invokeReductions(phase().getToken(), this, r);
    }

    // GraphViz //////////////////////////////////////////////////////////////////////////////

    public GraphViz.Node toGraphViz(GraphViz gv) {
        if (results.size()==0) return null;
        if (gv.hasNode(this)) return gv.createNode(this);
        GraphViz.Node n = gv.createNode(this);
        n.label = "state["+state.toInt()+"]";
        n.shape = "rectangle";
        boolean hasparents = false;
        for(Result r : results) n.edge(r, "");
        n.color = state.doomed ? "red" : "green";
        ((GraphViz.Group)phase().toGraphViz(gv)).add(n);
        return n;
    }
    public boolean isTransparent() { return false; }
    public boolean isHidden() { return false; }

}
